/* GDM - The GNOME Display Manager
 * Copyright (C) 1998, 1999, 2000 Martin K. Petersen <mkp@mkp.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef GREETER_CONFIGURATION_H
#define GREETER_CONFIGURATION_H

extern gboolean GdmUseCirclesInEntry;
extern gboolean GdmUseInvisibleInEntry;
extern gboolean GdmShowGnomeFailsafeSession;
extern gboolean GdmShowXtermFailsafeSession;
extern gboolean GdmShowLastSession;
extern gboolean GdmSystemMenu;
extern gboolean GdmConfigAvailable;
extern gboolean GdmChooserButton;
extern gchar *GdmHalt;
extern gchar *GdmReboot;
extern gchar *GdmSuspend;
extern gchar *GdmConfigurator;
extern gboolean GdmHaltFound;
extern gboolean GdmRebootFound;
extern gboolean GdmSuspendFound;
extern gboolean GdmConfiguratorFound;
extern gchar *GdmSessionDir;
extern gchar *GdmDefaultSession;
extern gchar *GdmDefaultLocale;
extern gchar *GdmLocaleFile;
extern gboolean GdmTimedLoginEnable;
extern gboolean GdmUse24Clock;
extern gchar *GdmTimedLogin;
extern gint GdmTimedLoginDelay;
extern gchar *GdmGlobalFaceDir;
extern gchar *GdmDefaultFace;
extern gint  GdmIconMaxHeight;
extern gint  GdmIconMaxWidth;
extern gchar *GdmExclude;
extern int GdmMinimalUID;
extern gboolean GdmAllowRoot;
extern gboolean GdmAllowRemoteRoot;
extern gchar *GdmWelcome;
extern gchar *GdmServAuthDir;
extern gchar *GdmInfoMsgFile;
extern gchar *GdmInfoMsgFont;
extern gchar *GdmSoundProgram;
extern gchar *GdmSoundOnLoginReadyFile;
extern gchar *GdmSoundOnLoginSuccessFile;
extern gchar *GdmSoundOnLoginFailureFile;
extern gboolean GdmSoundOnLoginReady;
extern gboolean GdmSoundOnLoginSuccess;
extern gboolean GdmSoundOnLoginFailure;

extern gboolean GDM_IS_LOCAL;
extern gboolean DOING_GDM_DEVELOPMENT;

#endif /* GREETER_CONFIGURATION_H */
